from .coefs import coefficients
from .operators import FinDiff, Coef, Identity, Coefficient
from .vector import Gradient, Divergence, Curl, Laplacian
from .pde import PDE, BoundaryConditions

__version__ = "0.8.6"
